/*
 * Copyright (c) 2024 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-Nordic-5-Clause
 */

/** @file
 * @brief Wi-Fi Softap sample
 */

#include <zephyr/logging/log.h>
LOG_MODULE_REGISTER(softap);

#include <zephyr/net/wifi_mgmt.h>
#include <zephyr/net/wifi_utils.h>
#include "net_private.h"
#include <zephyr/net/dhcpv4_server.h>

//http server
#include <stdio.h>
#include <errno.h>

#include <zephyr/kernel.h>
#include <zephyr/logging/log_ctrl.h>
#include <zephyr/net/net_ip.h>
#include <zephyr/net/socket.h>
#include <zephyr/sys/reboot.h>

#include <zephyr/net/net_config.h>
#include <zephyr/net/net_mgmt.h>
#include <zephyr/net/net_event.h>
#include <zephyr/net/conn_mgr_monitor.h>
#include <zephyr/net/conn_mgr_connectivity.h>

#include <zephyr/net/http/client.h>
#include <zephyr/net/http/parser.h>

#include <zephyr/net/wifi.h>
#include <zephyr/net/wifi_mgmt.h>

//uart
#include "uart_async_adapter.h"
#include <zephyr/drivers/uart.h>
#include <zephyr/usb/usb_device.h>

#include <zephyr/device.h>
#include <zephyr/devicetree.h>
#include <soc.h>
#include <zephyr/settings/settings.h>

#include <zephyr/drivers/gpio.h>
#include <cType.h>
#include <zephyr/sys/printk.h>
#include <zephyr/sys/byteorder.h>
#include <math.h>
#include <time.h>

#define CONTAINER_OF(ptr, type, field) ((type *)(((char *)(ptr)) - offsetof(type, field)))

#define WIFI_SAP_MGMT_EVENTS (NET_EVENT_WIFI_AP_ENABLE_RESULT)

// /* Macro used to subscribe to specific Zephyr NET management events. */
#define L4_EVENT_MASK (NET_EVENT_L4_CONNECTED | NET_EVENT_L4_DISCONNECTED)
#define CONN_LAYER_EVENT_MASK (NET_EVENT_CONN_IF_FATAL_ERROR)

#define SERVER_PORT			CONFIG_HTTP_SERVER_SAMPLE_PORT
#define MAX_CLIENT_QUEUE		6
#define STACK_SIZE			CONFIG_HTTP_SERVER_SAMPLE_STACK_SIZE
#define THREAD_PRIORITY			K_PRIO_COOP(CONFIG_NUM_COOP_PRIORITIES - 1)

/* Macro called upon a fatal error, reboots the device. */
#define FATAL_ERROR()								\
	LOG_ERR("Fatal error!%s", IS_ENABLED(CONFIG_RESET_ON_FATAL_ERROR) ?	\
				  " Rebooting the device" : "");		\
	LOG_PANIC();								\
	IF_ENABLED(CONFIG_REBOOT, (sys_reboot(0)))

// uart macros
#define UART_BUF_SIZE 813
#define UART_WAIT_FOR_BUF_DELAY K_MSEC(50)
#define UART_WAIT_FOR_RX 50000

static struct net_mgmt_event_callback wifi_sap_mgmt_cb;

/* Zephyr NET management event callback structures. */
static struct net_mgmt_event_callback l4_cb;
static struct net_mgmt_event_callback conn_cb;

static K_MUTEX_DEFINE(wifi_ap_sta_list_lock);
struct wifi_ap_sta_node {
	bool valid;
	struct wifi_ap_sta_info sta_info;
};
static struct wifi_ap_sta_node sta_list[CONFIG_SOFTAP_SAMPLE_MAX_STATIONS];


struct http_req {
	struct http_parser parser;
	int socket;
	bool received_all;
	enum http_method method;
	const char *url;
	size_t url_len;
	const char *body;
	size_t body_len;
};

struct http_resp {
    char *status_line;
    char *content_type;
    char *body;
    size_t body_length;
};

/* Forward declarations */
static void process_tcp4(void);

#define RESPONSE_200 "HTTP/1.1 200 OK\r\n"
#define RESPONSE_400 "HTTP/1.1 400 Bad Request\r\n\r\n"
#define RESPONSE_403 "HTTP/1.1 403 Forbidden\r\n\r\n"
#define RESPONSE_404 "HTTP/1.1 404 Not Found\r\n\r\n"
#define RESPONSE_405 "HTTP/1.1 405 Method Not Allowed\r\n\r\n"
#define RESPONSE_500 "HTTP/1.1 500 Internal Server Error\r\n\r\n"

/* Processing threads for incoming connections */
K_THREAD_STACK_ARRAY_DEFINE(tcp4_handler_stack, MAX_CLIENT_QUEUE, STACK_SIZE);
static struct k_thread tcp4_handler_thread[MAX_CLIENT_QUEUE];
static k_tid_t tcp4_handler_tid[MAX_CLIENT_QUEUE];
K_THREAD_DEFINE(tcp4_thread_id, STACK_SIZE,
		process_tcp4, NULL, NULL, NULL,
		THREAD_PRIORITY, 0, -1);

#define PERIODIC_STACK_SIZE 1024
#define PERIODIC_THREAD_PRIORITY 15
K_THREAD_STACK_DEFINE(periodic_stack, PERIODIC_STACK_SIZE);
struct k_thread periodic_thread_data;

static K_SEM_DEFINE(network_connected_sem, 0, 1);

static K_SEM_DEFINE(socket_sem, 0, 1);

static int tcp4_sock;
static int tcp4_accepted[MAX_CLIENT_QUEUE];

static struct http_parser_settings parser_settings;

// structures of uart
static const struct device *uart = DEVICE_DT_GET(DT_NODELABEL(uart0));				//*uart is responsible to handle uart api
static struct k_work_delayable uart_work;

struct uart_data_t {
	void *fifo_reserved;
	uint8_t data[UART_BUF_SIZE];
	uint16_t len;
};

typedef struct {
    int prodotem;
	int prodohumi;
	int prodonoi;
	int prodop1;
	int prodop10;
	int prodop25;
	int prodoco2;
	float prodohch;
	int prodovo;
	int prodoco;
	int prodored;
	int prodogreen;
	int prodoblue;
	float prodoUVa;
	float prodoUVb;
	float prodoUVc;
	int prodoir;
} SensorData;

// functions of uart
bool dataReceived;
// char *data_over_wifi = NULL;
struct uart_data_t *buf;
char data_wifi[275];

static time_t m_time, m_last_calibrate_time;
struct tm *local_time_now;

void nrf_cal_set_time(time_t epoch)
{
    time_t uncal_difftime, difftime, newtime;
    newtime = epoch;
    // LOG_INF("newtime epoch %ld",newtime);    
    m_last_calibrate_time = k_uptime_get();         //Kernel time at the time of epoch write
    m_time = newtime;
    // LOG_INF(" cal set time: %d,%d,%d\n",m_time,m_last_calibrate_time,newtime);
}    

time_t nrf_cal_get_time(void)
{
    time_t return_time;
    return_time = m_time + ((k_uptime_get()-m_last_calibrate_time))/1000;

	return return_time;
	
}

time_t nrf_cal_get_time_calibrated(void)
{
    return nrf_cal_get_time();
}

static void uart_cb(const struct device *dev, struct uart_event *evt, void *user_data)
{
	ARG_UNUSED(dev);									//Mark the unused parameter to avoid compiler warnings

	static size_t aborted_len;
	
	static uint8_t *aborted_buf;
	static bool disable_req = false;

	// below mentioned cases are the uart events
	switch (evt->type) 
	{

	case UART_RX_RDY:														//Some data was received and receive timeout occurred (if RX timeout is enabled) or when the receive buffer is full
		LOG_INF("UART_RX_RDY");
		buf = CONTAINER_OF(evt->data.rx.buf, struct uart_data_t, data);
		buf->len += evt->data.rx.len;

		  //LOG_INF("data len:%d",evt->data.rx.len);
		//   dataReceived = true;

		if (disable_req) {
			return;
		}
		
		
		// if ((evt->data.rx.buf[buf->len - 2] == ']') ||
		//     (evt->data.rx.buf[buf->len - 1] == ']')) {
		if(buf->len < 814){
			
			dataReceived = true;
			buf->data[buf->len] = '\0';
			new_process_uart_data(buf->data);

			// if (server_connect() >= 0) {
				// client_get_new_id();
			// }

			// count_pkt+=1;

			memset(buf->data, 0, sizeof(buf->data));
			buf->len = 0;

			// const char *filename = "/SD:/testfile.txt";
			// write_file(filename, data_wifi);
			// read_file(filename);

			disable_req = true;
			uart_rx_disable(uart);
			// buf->len = 0;
			
		}
		break;

	case UART_RX_DISABLED:													//This event is generated whenever receiver has been stopped, disabled or finished its operation (receive buffer filled) and can be enabled again using uart_rx_enable()
		LOG_INF("UART_RX_DISABLED");
		disable_req = false;

		
		// memset(data_wifi, '0', sizeof(data_wifi));
		// data_wifi[0] = 'H';

		// k_work_submit(&storing_func_work);
		// k_sem_give(&storing_func_sem_2);

		// k_sem_give(&storing_func_sem);

		buf = k_malloc(sizeof(*buf));
		if (buf) {
			buf->len = 0;
		} else {
			dataReceived = false;
			// k_free(buf);
			// LOG_WRN("Not able to allocate UART receive buffer");
			k_work_reschedule(&uart_work, UART_WAIT_FOR_BUF_DELAY);
			return;
		}

		

		uart_rx_enable(uart, buf->data, sizeof(buf->data),
			       UART_WAIT_FOR_RX);

		break;

	case UART_RX_BUF_REQUEST:												//Driver requests next buffer for continuous reception
		LOG_INF("UART_RX_BUF_REQUEST");
		buf = k_malloc(sizeof(*buf));
		// buf_temp = k_malloc(sizeof(*buf_temp));
		if (buf) {
			buf->len = 0;
			if(!uart_rx_buf_rsp(uart, buf->data, sizeof(buf->data)))			//sets the new buffer 
			{
				LOG_INF("New buffer set successfully");
			}
			else
			{
				LOG_ERR("There is error: %d",errno);
			}
		} else {
			// LOG_WRN("Not able to allocate UART receive buffer");
		}

		break;

	case UART_RX_BUF_RELEASED:												//Buffer is no longer used by UART driver
		LOG_INF("UART_RX_BUF_RELEASED\n");
		// LOG_INF("LENGTH OF buffer coming: %d",buf->len);
		buf = CONTAINER_OF(evt->data.rx_buf.buf, struct uart_data_t,
				   data);

		
		// if (buf->len > 0) 
		// {
		// 	k_fifo_put(&fifo_uart_rx_data, buf);
		// } 
		// else 
		// {
			k_free(buf);
			LOG_INF("Buffer is released");
		// }
		
		break;

	case UART_TX_ABORTED:												//Transmitting aborted due to timeout or uart_tx_abort() call
		LOG_INF("UART_TX_ABORTED");
		if (!aborted_buf) {
			aborted_buf = (uint8_t *)evt->data.tx.buf;
		}

		aborted_len += evt->data.tx.len;
		buf = CONTAINER_OF(aborted_buf, struct uart_data_t,
				   data);
		uart_tx(uart, &buf->data[aborted_len],
			buf->len - aborted_len, SYS_FOREVER_MS);

		break;
	
	case UART_RX_STOPPED:
        LOG_ERR("UART RX stopped, reason: %d", evt->data.rx_stop.reason);
		// uart_rx_disable(uart);
        uart_rx_enable(uart, buf->data, sizeof(buf->data), 50);
        break;

	default:
		break;
	}
}

static void uart_work_handler(struct k_work *item)
{
	struct uart_data_t *buf;

	buf = k_malloc(sizeof(*buf));
	if (buf) {
		buf->len = 0;
	} else {
		dataReceived = false;
		// LOG_WRN("Not able to allocate UART receive buffer");
		k_work_reschedule(&uart_work, UART_WAIT_FOR_BUF_DELAY);						//Reschedules the delayable work item to be executed after a specified delay.
		return;
	}
	uart_rx_enable(uart, buf->data, sizeof(buf->data), UART_WAIT_FOR_RX);
}

// static bool uart_test_async_api(const struct device *dev)
// {
// 	const struct uart_driver_api *api =
// 			(const struct uart_driver_api *)dev->api;

// 	return (api->callback_set != NULL);
// }

static int uart_init(void)
{
	int err;
	int pos;
	struct uart_data_t *rx;

	if (!device_is_ready(uart)) {
		return -ENODEV;
	}

	if (IS_ENABLED(CONFIG_USB_DEVICE_STACK)) {
		err = usb_enable(NULL);
		if (err && (err != -EALREADY)) {
			LOG_ERR("Failed to enable USB");
			return err;
		}
	}

	rx = k_malloc(sizeof(*rx));
	if (rx) {
		rx->len = 0;
	} else {
		return -ENOMEM;
	}

	k_work_init_delayable(&uart_work, uart_work_handler);			//initialises a delayable work item which will be executed after a specific delay, This function prepares the work item to be added to a workqueue for execution


	//registering the callback function, uart_cb is the event handler which will handle the callback functionality
	err = uart_callback_set(uart, uart_cb, NULL);
	if (err) {
		k_free(rx);
		LOG_ERR("Cannot initialize UART callback");
		return err;
	}

	if (IS_ENABLED(CONFIG_UART_LINE_CTRL)) {
		LOG_INF("Wait for DTR");
		while (true) {
			uint32_t dtr = 0;

			uart_line_ctrl_get(uart, UART_LINE_CTRL_DTR, &dtr);
			if (dtr) {
				break;
			}
			/* Give CPU resources to low priority threads. */
			k_sleep(K_MSEC(100));
		}
		LOG_INF("DTR set");
		err = uart_line_ctrl_set(uart, UART_LINE_CTRL_DCD, 1);
		if (err) {
			LOG_WRN("Failed to set DCD, ret code %d", err);
		}
		err = uart_line_ctrl_set(uart, UART_LINE_CTRL_DSR, 1);
		if (err) {
			LOG_WRN("Failed to set DSR, ret code %d", err);
		}
	}
	
	// memset(rx->data, 0, sizeof(rx->data));

	// uart_rx_disable(uart);

	//to start receiving we will calll the below func
	err = uart_rx_enable(uart, rx->data, sizeof(rx->data), 50);
	if (err) {
		LOG_ERR("Cannot enable uart reception (err: %d)", err);
		/* Free the rx buffer only because the tx buffer will be handled in the callback */
		k_free(rx);
	}
	

	return err;
}

int parse_value(const char *str) {
    int value = 0;
    // Assuming that we find the key, skip the "value:" substring which is 8 characters long
    str += 8; // Skip over "\"value\":"
    // Skip any spaces or additional characters until the start of the number
    while (*str < '0' || *str > '9') {
        str++;
    }
    // Convert the numerical characters to an integer
    while (*str >= '0' && *str <= '9') {
        value = value * 10 + (*str - '0');
        str++;
    }
    return value;
}

void new_process_uart_data(const char *uart_data) {
	SensorData *sensor;
	

	
// *ppm =25;
	sensor = (SensorData *)k_malloc(sizeof(SensorData));
    if (sensor == NULL) {
        // Handle memory allocation failure
        printf("Memory allocation failed\n");
        return 1;
    }
	int tem = 0, humi = 0, noi = 0, p10 = 0, p1 = 0, p25 = 0, co2 = 0, hch = 0, vo = 0, co = 0, red = 0, blue = 0, green = 0, uva = 0, uvb = 0, uvc = 0, ir = 0;
    // Extract temperature

		char *tempPtr = strstr(uart_data, "Temp");
		// LOG_INF("Temp ptr value: %s",tempPtr);
		if (tempPtr) {
			char *valuePtr = strstr(tempPtr, "\"value\":\t");
			// LOG_INF("Temp value: %s",valuePtr);
			if (valuePtr) {
				tem = parse_value(valuePtr);
			}
		}
    // Extract humidity
		char *humPtr = strstr(uart_data, "Hum");
		if (humPtr) {
			char *valuePtr = strstr(humPtr, "\"value\":\t");
			if (valuePtr) {
				humi = parse_value(valuePtr);
			}
		}
    // Extract noise
		char *noisePtr = strstr(uart_data, "Noi-DB");
		if (noisePtr) {
			char *valuePtr = strstr(noisePtr, "\"value\":\t");
			if (valuePtr) {
				noi = parse_value(valuePtr);
			}
		}

		char *ppm1Ptr = strstr(uart_data, "Pm1.0");
		if (ppm1Ptr) {
			char *valuePtr = strstr(ppm1Ptr, "\"value\":\t");
			if (valuePtr) {
				p1 = parse_value(valuePtr);
			}
		}

		char *ppm10Ptr = strstr(uart_data, "Pm10");
		if (ppm10Ptr) {
			char *valuePtr = strstr(ppm10Ptr, "\"value\":\t");
			if (valuePtr) {
				p10 = parse_value(valuePtr);
			}
		}

		char *ppm25Ptr = strstr(uart_data, "Pm25");
		if (ppm25Ptr) {
			char *valuePtr = strstr(ppm25Ptr, "\"value\":\t");
			if (valuePtr) {
				p25 = parse_value(valuePtr);
			}
		}

		char *co_2ptr = strstr(uart_data, "Co2");
		if (co_2ptr) {
			char *valuePtr = strstr(co_2ptr, "\"value\":\t");
			if (valuePtr) {
				co2 = parse_value(valuePtr);
			}
		}

		char *hcho_ptr = strstr(uart_data, "Hcho");
		if (hcho_ptr) {
			char *valuePtr = strstr(hcho_ptr, "\"value\":\t");
			if (valuePtr) {
				hch = parse_value(valuePtr);
			}
		}

		char *vocptr = strstr(uart_data, "Voc");
		if (vocptr) {
			char *valuePtr = strstr(vocptr, "\"value\":\t");
			if (valuePtr) {
				vo = parse_value(valuePtr);
			}
		}

		char *coptr = strstr(uart_data, "Co");
		if (coptr) {
			char *valuePtr = strstr(coptr, "\"value\":\t");
			if (valuePtr) {
				co = parse_value(valuePtr);
			}
		}

		char *redptr = strstr(uart_data, "RED");
		if (redptr) {
			char *valuePtr = strstr(redptr, "\"value\":\t");
			if (valuePtr) {
				red = parse_value(valuePtr);
			}
		}

		char *greenptr = strstr(uart_data, "Green");
		if (greenptr) {
			char *valuePtr = strstr(greenptr, "\"value\":\t");
			if (valuePtr) {
				green = parse_value(valuePtr);
			}
		}

		char *blueptr = strstr(uart_data, "Bulle");
		if (blueptr) {
			char *valuePtr = strstr(blueptr, "\"value\":\t");
			if (valuePtr) {
				blue = parse_value(valuePtr);
			}
		}

		char *irptr = strstr(uart_data, "Ir");
		if (irptr) {
			char *valuePtr = strstr(irptr, "\"value\":\t");
			if (valuePtr) {
				ir = parse_value(valuePtr);
			}
		}

		char *uvaptr = strstr(uart_data, "UVA");
		if (uvaptr) {
			char *valuePtr = strstr(uvaptr, "\"value\":\t");
			if (valuePtr) {
				uva = parse_value(valuePtr);
			}
		}

		char *uvbptr = strstr(uart_data, "UVB");
		if (uvbptr) {
			char *valuePtr = strstr(uvbptr, "\"value\":\t");
			if (valuePtr) {
				red = parse_value(valuePtr);
			}
		}

		char *uvcptr = strstr(uart_data, "UVC");
		if (uvcptr) {
			char *valuePtr = strstr(uvcptr, "\"value\":\t");
			if (valuePtr) {
				red = parse_value(valuePtr);
			}
		}
		sensor->prodotem = tem;
		sensor->prodohumi = humi;
		sensor->prodonoi = noi;
		sensor->prodop1 = p1;
		sensor->prodop10 = p10;
		sensor->prodop25 = p25;
		sensor->prodoco2 =co2;
		sensor->prodohch = hch;
		sensor->prodovo = vo;
		sensor->prodoco = co;
		sensor->prodored = red;
		sensor->prodoblue = blue;
		sensor->prodogreen = green;
		sensor->prodoUVa = uva;
		sensor->prodoUVb = uvb;
		sensor->prodoUVc = uvc;
		sensor->prodoir = ir;

		if(tem!=0 && humi!=0){
			snprintf(data_wifi, sizeof(data_wifi),"{\"Hcho-mg/m3\": %.2f,\"Co-PPM\": %d,\"Co2-PPM\": %d,\"Pm1.0-¦Ìg/m3\": %d,\"Pm10-¦Ìg/m3\": %d,\"Pm25-¦Ìg/m3\": %d,\"Voc-mg/m3\": %d,\"Temp-¡æ\": %d,\"Hum-\": %d,\"Noi-DB\": %d,\"UVA-mw/cm2\": %.2f,\"UVB-mw/cm2\": %.2f,\"UVC-mw/cm2\": %.2f,\"RED\": %d,\"GREEN\": %d,\"BLUE\": %d,\"IR\": %d}",
				(float)sensor->prodohch,
				sensor->prodoco,
				sensor->prodoco2,
				sensor->prodop1,
				sensor->prodop10,
				sensor->prodop25,
				sensor->prodovo,
				sensor->prodotem,
				sensor->prodohumi,
				sensor->prodonoi,
				(sensor->prodoUVa/100),
				(sensor->prodoUVb/100),
				(sensor->prodoUVc/100),
				sensor->prodored,
				sensor->prodogreen,
				sensor->prodoblue,
				sensor->prodoir);

			time_t newtime = nrf_cal_get_time_calibrated();
		
		
			// // LOG_INF("m_last_calibrate_time: %ld",m_last_calibrate_time);
			// LOG_INF("Newtime: %ld",(long)newtime);
			// k_sleep(K_SECONDS(1));
			local_time_now= localtime(&newtime);
			char current_time[9];
			snprintf(current_time, sizeof(current_time), "%02d:%02d:%02d",local_time_now->tm_hour,local_time_now->tm_min,local_time_now->tm_sec);
			memmove(data_wifi + strlen(current_time) + 1, data_wifi, strlen(data_wifi) + 1);
			memcpy(data_wifi, current_time, strlen(current_time));
			data_wifi[strlen(current_time)] = ' ';

			
			LOG_INF("The data from prodo is: %s\r\n",data_wifi);

		}

		// strcpy(server_data,data_wifi);
		
		// client_get_new_id();
		// if (client_get_new_id() < 0) {
		// 	LOG_ERR("Failed to get client ID");
		// 	return 0;
		// }
		k_free(sensor);

		// LOG_INF("Size of data wifi string: %d",strlen(data_wifi));
		// write_file(filename, data_wifi);
		
}	

static void l4_event_handler(struct net_mgmt_event_callback *cb,
			     uint32_t event,
			     struct net_if *iface)
{
	switch (event) {
	case NET_EVENT_L4_CONNECTED:
		LOG_INF("Network connected");

		k_sem_give(&network_connected_sem);
		break;
	case NET_EVENT_L4_DISCONNECTED:
		LOG_INF("Network disconnected");
		break;
	default:
		/* Don't care */
		return;
	}
}

static void connectivity_event_handler(struct net_mgmt_event_callback *cb,
				       uint32_t event,
				       struct net_if *iface)
{
	if (event == NET_EVENT_CONN_IF_FATAL_ERROR) {
		LOG_ERR("NET_EVENT_CONN_IF_FATAL_ERROR");
		FATAL_ERROR();
		return;
	}
}



static int build_response(struct http_resp *resp, char *buffer, size_t buffer_size) {
    int ret;

    ret = snprintk(buffer, buffer_size,
                   "%sContent-Type: %s\r\nContent-Length: %zu\r\n\r\n%s",
                   resp->status_line, resp->content_type, resp->body_length, resp->body);
    if ((ret < 0) || (ret >= buffer_size)) {
        return -ENOBUFS;
    }

    return 0;
}

static int send_response(struct http_req *request, char *response)
{
	
	size_t len = strlen(response);

	while (len) {
		ssize_t out_len;
		out_len = send(request->socket, response, len, 0);
		if (out_len < 0) {
			LOG_ERR("send, error: %d", -errno);
			return -errno;
		}
		response += out_len;
        len -= out_len;
	}

	return 0;
}

static void send_periodic_response(void *arg1, void *arg2, void *arg3)
{
    int client_socket = *(int *)arg1;
    struct http_resp resp;
    char dynamic_response_buf[500];
    int ret;

    while (true) {


        // Check if the socket is still valid
        if (client_socket < 0) {
            break;
        }

        resp.status_line = "HTTP/1.1 200 OK\r\n";
        resp.content_type = "text/plain";
        resp.body = "Hello";
        resp.body_length = strlen(resp.body);

        ret = build_response(&resp, dynamic_response_buf, sizeof(dynamic_response_buf));
        if (ret == 0) {
            size_t len = strlen(dynamic_response_buf);
            size_t sent = 0;

            while (sent < len) {
                ssize_t out_len = send(client_socket, dynamic_response_buf + sent, len - sent, 0);
                if (out_len < 0) {
                    LOG_ERR("send, error: %d", -errno);
                    close(client_socket);
                    client_socket = -1;
                    return;
                }
                sent += out_len;
            }
        } else {
            LOG_ERR("Failed to build response, error: %d", ret);
            close(client_socket);
            client_socket = -1;
            return;
        }

        // k_sem_give(&socket_sem); // Give the semaphore
        k_sleep(K_SECONDS(1)); // Adjust the sleep duration as needed
    }
}

/* Handle HTTP request */
static void handle_http_request(struct http_req *request)
{
	// int ret;
	// char *resp_ptr = RESPONSE_200;
	// char get_response_buffer[100] = { 0 };

	int ret;
	struct http_resp resp;
    
	size_t len;
	const char *resp_ptr = RESPONSE_403;
	char dynamic_response_buf[500];
	char url[100];
	// size_t url_len = MIN(sizeof(url) - 1, request->url_len);
	size_t url_len = request->url_len;

	memcpy(url, request->url, url_len);

	url[url_len] = '\0';

	/* Handle the request method */
	switch (request->method) {
	case HTTP_GET:
		LOG_INF("GET %s", request->url);
		if (strcmp(url, "/data") == 0) {
			resp.status_line = "HTTP/1.1 200 OK\r\n";
			resp.content_type = "text/plain";
			resp.body = "Hello!!!";
			resp.body_length = strlen(resp.body);
			// while (true) {
			// 	resp.body = "Hello";
			// 	resp.body_length = strlen(resp.body);

			// 	ret = build_response(&resp, dynamic_response_buf, sizeof(dynamic_response_buf));
			// 	if (ret == 0) {
			// 		send_response(request, dynamic_response_buf);
			// 	} else {
			// 		LOG_ERR("Failed to build response, error: %d", ret);
			// 		break;
			// 	}
			// 	k_sleep(K_SECONDS(1)); // Adjust the sleep duration as needed
			// }
			// static int client_socket = -1;
			// client_socket = request->socket;
            // k_thread_create(&periodic_thread_data, periodic_stack,
            //                 K_THREAD_STACK_SIZEOF(periodic_stack),
            //                 send_periodic_response,
            //                 &client_socket, NULL, NULL,
            //                 PERIODIC_THREAD_PRIORITY, 0, K_NO_WAIT);
            // return;
		} else {
			resp.status_line = "HTTP/1.1 404 Not Found\r\n";
			resp.content_type = "text/plain";
			resp.body = "Resource not found";
			resp.body_length = strlen(resp.body);

			// ret = build_response(&resp, dynamic_response_buf, sizeof(dynamic_response_buf));
			// if (ret == 0) {
			// 	send_response(request, dynamic_response_buf);
			// } else {
			// 	LOG_ERR("Failed to build response, error: %d", ret);
			// 	FATAL_ERROR();
			// }
		}
		break;


	default:
            LOG_WRN("Unsupported request method");
            resp.status_line = "HTTP/1.1 405 Method Not Allowed\r\n";
            resp.content_type = "text/plain";
            resp.body = "Method not allowed";
            resp.body_length = strlen(resp.body);

			// ret = build_response(&resp, dynamic_response_buf, sizeof(dynamic_response_buf));
			// if (ret == 0) {
			// 	send_response(request, dynamic_response_buf);
			// } else {
			// 	LOG_ERR("Failed to build response, error: %d", ret);
			// 	FATAL_ERROR();
			// }
            break;
			
	}

	ret = build_response(&resp, dynamic_response_buf, sizeof(dynamic_response_buf));
    if (ret == 0) {
        send_response(request, dynamic_response_buf);
    } else {
        LOG_ERR("Failed to build response, error: %d", ret);
        FATAL_ERROR();
    }
	
}

/* HTTP parser callbacks */
static int on_body(struct http_parser *parser, const char *at, size_t length)
{
	struct http_req *req = CONTAINER_OF(parser, struct http_req, parser);

	req->body = at;
	req->body_len = length;

	LOG_DBG("on_body: %d", parser->method);
	LOG_DBG("> %.*s", length, at);

	return 0;
}

static int on_headers_complete(struct http_parser *parser)
{
	struct http_req *req = CONTAINER_OF(parser, struct http_req, parser);

	req->method = parser->method;

	LOG_DBG("on_headers_complete, method: %s", http_method_str(parser->method));

	return 0;
}

static int on_message_begin(struct http_parser *parser)
{
	struct http_req *req = CONTAINER_OF(parser, struct http_req, parser);

	req->received_all = false;

	LOG_DBG("on_message_begin, method: %d", parser->method);

	return 0;
}

static int on_message_complete(struct http_parser *parser)
{
	struct http_req *req = CONTAINER_OF(parser, struct http_req, parser);

	req->received_all = true;

	LOG_DBG("on_message_complete, method: %d", parser->method);

	return 0;
}

static int on_url(struct http_parser *parser, const char *at, size_t length)
{
	struct http_req *req = CONTAINER_OF(parser, struct http_req, parser);

	req->url = at;
	req->url_len = length;

	LOG_DBG("on_url, method: %d", parser->method);
	LOG_DBG("> %.*s", length, at);

	return 0;
}

/* Initialize HTTP parser. */
static void parser_init(void)
{
	http_parser_settings_init(&parser_settings);

	parser_settings.on_body = on_body;
	parser_settings.on_headers_complete = on_headers_complete;
	parser_settings.on_message_begin = on_message_begin;
	parser_settings.on_message_complete = on_message_complete;
	parser_settings.on_url = on_url;
}

static int setup_server(int *sock, struct sockaddr *bind_addr, socklen_t bind_addrlen)
{
	int ret;

	*sock = socket(bind_addr->sa_family, SOCK_STREAM, IPPROTO_TCP);

	if (*sock < 0) {
		LOG_ERR("Failed to create TCP socket: %d", errno);
		return -errno;
	}

	ret = bind(*sock, bind_addr, bind_addrlen);
	if (ret < 0) {
		LOG_ERR("Failed to bind TCP socket %d", errno);
		return -errno;
	}

	ret = listen(*sock, MAX_CLIENT_QUEUE);
	if (ret < 0) {
		LOG_ERR("Failed to listen on TCP socket %d", errno);
		ret = -errno;
	}

	return ret;
}

static void client_conn_handler(void *ptr1, void *ptr2, void *ptr3)
{
	ARG_UNUSED(ptr1);
	int *sock = ptr2;
	k_tid_t *in_use = ptr3;
	int received;
	int ret;
	char buf[CONFIG_HTTP_SERVER_SAMPLE_RECEIVE_BUFFER_SIZE];
	size_t offset = 0;
	size_t total_received = 0;
	struct http_req request = {
		.socket = *sock,
	};

	http_parser_init(&request.parser, HTTP_REQUEST);

	while (true) {
		received = recv(request.socket, buf + offset, sizeof(buf) - offset, 0);
		if (received == 0) {
			/* Connection closed */
			LOG_DBG("[%d] Connection closed by peer", request.socket);
			break;
		} else if (received < 0) {
			/* Socket error */
			ret = -errno;
			LOG_ERR("[%d] Connection error %d", request.socket, ret);
			break;
		}

		/* Parse the received data as HTTP request */
		(void)http_parser_execute(&request.parser,
					  &parser_settings, buf + offset, received);

		total_received += received;
		offset += received;

		if (offset >= sizeof(buf)) {
			offset = 0;
		}

		/* If the HTTP request has been completely received, stop receiving data and
		 * proceed to process the request.
		 */
		if (request.received_all) {
			handle_http_request(&request);
			break;
		}
	};

	(void)close(request.socket);

	*sock = -1;
	*in_use = NULL;
}

static int get_free_slot(int *accepted)
{
	for (int i = 0; i < MAX_CLIENT_QUEUE; i++) {
		if (accepted[i] < 0) {
			return i;
		}
	}

	return -1;
}

static void process_tcp(sa_family_t family, int *sock, int *accepted)
{
	static int counter;
	int client;
	int slot;
	struct sockaddr_in client_addr;
	socklen_t client_addr_len = sizeof(client_addr);
	char addr_str[INET_ADDRSTRLEN];

	client = accept(*sock, (struct sockaddr *)&client_addr,
			&client_addr_len);
	if (client < 0) {
		LOG_ERR("Error in accept %d, stopping server", -errno);
		return -errno;
	}
	else
	{
		LOG_INF("Accepted the socket");
	}

	slot = get_free_slot(accepted);
	if (slot < 0 || slot >= MAX_CLIENT_QUEUE) {
		LOG_ERR("Cannot accept more connections");
		close(client);
		return 0;
	}

	accepted[slot] = client;

	if (client_addr.sin_family == AF_INET) {
		tcp4_handler_tid[slot] = k_thread_create(
			&tcp4_handler_thread[slot],
			tcp4_handler_stack[slot],
			K_THREAD_STACK_SIZEOF(tcp4_handler_stack[slot]),
			(k_thread_entry_t)client_conn_handler,
			INT_TO_POINTER(slot),
			&accepted[slot],
			&tcp4_handler_tid[slot],
			THREAD_PRIORITY,
			0, K_NO_WAIT);
	}

	net_addr_ntop(client_addr.sin_family, &client_addr.sin_addr, addr_str, sizeof(addr_str));

	LOG_INF("[%d] Connection #%d from %s", client, ++counter, addr_str);
    

}

/* Processing incoming IPv4 clients */
static void process_tcp4(void)
{
	int ret;
	struct sockaddr_in addr4 = {
		.sin_family = AF_INET,
		.sin_port = htons(SERVER_PORT),
	};

	ret = setup_server(&tcp4_sock, (struct sockaddr *)&addr4, sizeof(addr4));
	if (ret < 0) {
		LOG_ERR("Failed to create IPv4 socket %d", ret);
		FATAL_ERROR();
		return;
	}

	LOG_INF("Waiting for IPv4 HTTP connections on port %d, sock %d", SERVER_PORT, tcp4_sock);

	
	while (true) {
		process_tcp(AF_INET, &tcp4_sock, tcp4_accepted);
	}
}


void start_listener(void)
{
	for (size_t i = 0; i < MAX_CLIENT_QUEUE; i++) {
		if (IS_ENABLED(CONFIG_NET_IPV4)) {
			tcp4_accepted[i] = -1;
			tcp4_sock = -1;
		}

	}

	


	if (IS_ENABLED(CONFIG_NET_IPV4)) {
		k_thread_start(tcp4_thread_id);
	}
}

static void wifi_ap_stations_unlocked(void)
{
	size_t id = 1;

	LOG_INF("AP stations:");
	LOG_INF("============");

	for (int i = 0; i < CONFIG_SOFTAP_SAMPLE_MAX_STATIONS; i++) {
		struct wifi_ap_sta_info *sta;
		uint8_t mac_string_buf[sizeof("xx:xx:xx:xx:xx:xx")];

		if (!sta_list[i].valid) {
			continue;
		}

		sta = &sta_list[i].sta_info;

		LOG_INF("Station %zu:", id++);
		LOG_INF("==========");
		LOG_INF("MAC: %s",
			net_sprint_ll_addr_buf(sta->mac,
					       WIFI_MAC_ADDR_LEN,
					       mac_string_buf,
					       sizeof(mac_string_buf)));
		LOG_INF("Link mode: %s", wifi_link_mode_txt(sta->link_mode));
		LOG_INF("TWT: %s", sta->twt_capable ? "Supported" : "Not supported");
	}

	if (id == 1) {
		LOG_INF("No stations connected");
	}
}

static void handle_wifi_ap_enable_result(struct net_mgmt_event_callback *cb)
{
	const struct wifi_status *status =
		(const struct wifi_status *)cb->info;

	if (status->status) {
		LOG_ERR("AP enable request failed (%d)", status->status);
	} else {
		LOG_INF("AP enable requested");
	}
}

static void handle_wifi_ap_sta_connected(struct net_mgmt_event_callback *cb)
{
	const struct wifi_ap_sta_info *sta_info =
		(const struct wifi_ap_sta_info *)cb->info;
	uint8_t mac_string_buf[sizeof("xx:xx:xx:xx:xx:xx")];
	int i;

	LOG_INF("Station connected: %s",
		net_sprint_ll_addr_buf(sta_info->mac, WIFI_MAC_ADDR_LEN,
				       mac_string_buf, sizeof(mac_string_buf)));

	k_mutex_lock(&wifi_ap_sta_list_lock, K_FOREVER);
	for (i = 0; i < CONFIG_SOFTAP_SAMPLE_MAX_STATIONS; i++) {
		if (!sta_list[i].valid) {
			sta_list[i].sta_info = *sta_info;
			sta_list[i].valid = true;
			break;
		}
	}

	if (i == CONFIG_SOFTAP_SAMPLE_MAX_STATIONS) {
		LOG_ERR("No space to store station info: "
			"Increase CONFIG_SOFTAP_SAMPLE_MAX_STATIONS");
	}

	wifi_ap_stations_unlocked();
	k_mutex_unlock(&wifi_ap_sta_list_lock);
}

static void handle_wifi_ap_sta_disconnected(struct net_mgmt_event_callback *cb)
{
	const struct wifi_ap_sta_info *sta_info =
		(const struct wifi_ap_sta_info *)cb->info;
	uint8_t mac_string_buf[sizeof("xx:xx:xx:xx:xx:xx")];
	int i;

	LOG_INF("Station disconnected: %s",
		net_sprint_ll_addr_buf(sta_info->mac, WIFI_MAC_ADDR_LEN,
				       mac_string_buf, sizeof(mac_string_buf)));

	k_mutex_lock(&wifi_ap_sta_list_lock, K_FOREVER);
	for (i = 0; i < CONFIG_SOFTAP_SAMPLE_MAX_STATIONS; i++) {
		if (!sta_list[i].valid) {
			continue;
		}

		if (!memcmp(sta_list[i].sta_info.mac, sta_info->mac,
			    WIFI_MAC_ADDR_LEN)) {
			sta_list[i].valid = false;
			break;
		}
	}

	if (i == CONFIG_SOFTAP_SAMPLE_MAX_STATIONS) {
		LOG_WRN("No matching MAC address found in the list");
	}

	wifi_ap_stations_unlocked();
	k_mutex_unlock(&wifi_ap_sta_list_lock);
}

static void wifi_mgmt_event_handler(struct net_mgmt_event_callback *cb,
				    uint32_t mgmt_event, struct net_if *iface)
{
	switch (mgmt_event) {
	case NET_EVENT_WIFI_AP_ENABLE_RESULT:
		handle_wifi_ap_enable_result(cb);
		break;
	case NET_EVENT_WIFI_AP_STA_CONNECTED:
		handle_wifi_ap_sta_connected(cb);
		break;
	case NET_EVENT_WIFI_AP_STA_DISCONNECTED:
		handle_wifi_ap_sta_disconnected(cb);
		break;
	default:
		break;
	}
}

static int __wifi_args_to_params(struct wifi_connect_req_params *params)
{
#ifdef CONFIG_SOFTAP_SAMPLE_2_4GHz
	params->band = WIFI_FREQ_BAND_2_4_GHZ;
#elif CONFIG_SOFTAP_SAMPLE_5GHz
	params->band = WIFI_FREQ_BAND_5_GHZ;
#endif
	params->channel = CONFIG_SOFTAP_SAMPLE_CHANNEL;

	/* SSID */
	params->ssid = CONFIG_SOFTAP_SAMPLE_SSID;
	params->ssid_length = strlen(params->ssid);
	if (params->ssid_length > WIFI_SSID_MAX_LEN) {
		LOG_ERR("SSID length is too long, expected is %d characters long",
			WIFI_SSID_MAX_LEN);
		return -1;
	}

#if defined(CONFIG_SOFTAP_SAMPLE_KEY_MGMT_WPA2)
	params->security = 1;
#elif defined(CONFIG_SOFTAP_SAMPLE_KEY_MGMT_WPA2_256)
	params->security = 2;
#elif defined(CONFIG_SOFTAP_SAMPLE_KEY_MGMT_WPA3)
	params->security = 3;
#else
	params->security = 0;
#endif

#if !defined(CONFIG_SOFTAP_SAMPLE_KEY_MGMT_NONE)
	params->psk = CONFIG_SOFTAP_SAMPLE_PASSWORD;
	params->psk_length = strlen(params->psk);
#endif

	return 0;
}

static int cmd_wifi_status(void)
{
	struct net_if *iface;
	struct wifi_iface_status status = { 0 };

	iface = net_if_get_first_wifi();
	if (!iface) {
		LOG_ERR("Failed to get Wi-FI interface");
		return -ENOEXEC;
	}

	if (net_mgmt(NET_REQUEST_WIFI_IFACE_STATUS, iface, &status,
				sizeof(struct wifi_iface_status))) {
		LOG_ERR("Status request failed");
		return -ENOEXEC;
	}

	LOG_INF("Status: successful");
	LOG_INF("==================");
	LOG_INF("State: %s", wifi_state_txt(status.state));

	if (status.state >= WIFI_STATE_ASSOCIATED) {
		uint8_t mac_string_buf[sizeof("xx:xx:xx:xx:xx:xx")];

		LOG_INF("Interface Mode: %s", wifi_mode_txt(status.iface_mode));
		LOG_INF("Link Mode: %s", wifi_link_mode_txt(status.link_mode));
		LOG_INF("SSID: %.32s", status.ssid);
		LOG_INF("BSSID: %s",
			net_sprint_ll_addr_buf(status.bssid,
					       WIFI_MAC_ADDR_LEN, mac_string_buf,
					       sizeof(mac_string_buf)));
		LOG_INF("Band: %s", wifi_band_txt(status.band));
		LOG_INF("Channel: %d", status.channel);
		LOG_INF("Security: %s", wifi_security_txt(status.security));
		LOG_INF("MFP: %s", wifi_mfp_txt(status.mfp));
		LOG_INF("Beacon Interval: %d", status.beacon_interval);
		LOG_INF("DTIM: %d", status.dtim_period);
		LOG_INF("TWT: %s",
			status.twt_capable ? "Supported" : "Not supported");
	}

	return 0;
}

static void wifi_softap_enable(void)
{
	struct net_if *iface;
	static struct wifi_connect_req_params cnx_params;
	int ret;

	iface = net_if_get_first_wifi();
	if (!iface) {
		LOG_ERR("Failed to get Wi-Fi iface");
		return;
	}

	if (__wifi_args_to_params(&cnx_params)) {
		return;
	}

	if (!wifi_utils_validate_chan(cnx_params.band, cnx_params.channel)) {
		LOG_ERR("Invalid channel %d in %d band",
			cnx_params.channel, cnx_params.band);
		return;
	}

	ret = net_mgmt(NET_REQUEST_WIFI_AP_ENABLE, iface, &cnx_params,
		       sizeof(struct wifi_connect_req_params));
	if (ret) {
		LOG_ERR("AP mode enable failed: %s", strerror(-ret));
	} else {
		LOG_INF("AP mode enabled");
	}
}

static void wifi_set_reg_domain(void)
{
	struct net_if *iface;
	struct wifi_reg_domain regd = {0};
	int ret;

	iface = net_if_get_first_wifi();
	if (!iface) {
		LOG_ERR("Failed to get Wi-Fi iface");
		return;
	}

	regd.oper = WIFI_MGMT_SET;
	strncpy(regd.country_code, CONFIG_SOFTAP_SAMPLE_REG_DOMAIN,
		(WIFI_COUNTRY_CODE_LEN + 1));

	ret = net_mgmt(NET_REQUEST_WIFI_REG_DOMAIN, iface,
		       &regd, sizeof(regd));
	if (ret) {
		LOG_ERR("Cannot %s Regulatory domain: %d", "SET", ret);
	} else {
		LOG_INF("Regulatory domain set to %s", CONFIG_SOFTAP_SAMPLE_REG_DOMAIN);
	}
}

static void configure_dhcp_server(void)
{
	struct net_if *iface;
	struct in_addr pool_start;
	int ret;

	iface = net_if_get_first_wifi();
	if (!iface) {
		LOG_ERR("Failed to get Wi-Fi interface");
		return;
	}

	if (net_addr_pton(AF_INET, CONFIG_SOFTAP_SAMPLE_DHCPV4_POOL_START, &pool_start.s_addr)) {
		LOG_ERR("Invalid address: %s", CONFIG_SOFTAP_SAMPLE_DHCPV4_POOL_START);
		return;
	}

	ret = net_dhcpv4_server_start(iface, &pool_start);
	if (ret == -EALREADY) {
		LOG_ERR("DHCPv4 server already running on interface");
	} else if (ret < 0) {
		LOG_ERR("DHCPv4 server failed to start and returned %d error", ret);
	} else {
		LOG_INF("DHCPv4 server started and pool address starts from %s",
			CONFIG_SOFTAP_SAMPLE_DHCPV4_POOL_START);
	}
}


int main(void)
{
	
	int ret;

	// struct net_if *iface;
	// iface = net_if_get_first_wifi();
    // if (!iface) {
    //     LOG_ERR("Failed to get Wi-Fi interface");
    //     return;
    // }
	

	parser_init();


	k_sleep(K_SECONDS(1));

	

	net_mgmt_init_event_callback(&wifi_sap_mgmt_cb,
				     wifi_mgmt_event_handler,
				     WIFI_SAP_MGMT_EVENTS);

	net_mgmt_add_event_callback(&wifi_sap_mgmt_cb);

	/* Setup handler for Zephyr NET Connection Manager events. */
	net_mgmt_init_event_callback(&l4_cb, l4_event_handler, L4_EVENT_MASK);
	net_mgmt_add_event_callback(&l4_cb);

	/* Setup handler for Zephyr NET Connection Manager Connectivity layer. */
	net_mgmt_init_event_callback(&conn_cb, connectivity_event_handler, CONN_LAYER_EVENT_MASK);
	net_mgmt_add_event_callback(&conn_cb);

	ret = conn_mgr_all_if_up(true);
	if (ret) {
		LOG_ERR("conn_mgr_all_if_up, error: %d", ret);
		// FATAL_ERROR();
		return ret;
	}

	LOG_INF("Network interface brought up");

	ret = conn_mgr_all_if_connect(true);
	if (ret) {
		LOG_ERR("conn_mgr_all_if_connect, error: %d", ret);
		// FATAL_ERROR();
		return ret;
	}



	wifi_set_reg_domain();

	configure_dhcp_server();

	wifi_softap_enable();

	cmd_wifi_status();

	k_sem_take(&network_connected_sem, K_FOREVER);

	

	start_listener();

	// ret = uart_init();
	// if (ret < 0) {
	// 	// error();
	// 	LOG_ERR("UART not initialised");
	// }
	// else
	// {
	// 	LOG_INF("UART initialised");
	// }

	return 0;
}
